from typing import Callable

import numpy as np

from sampo.scheduler.resource.base import ResourceOptimizer
from sampo.scheduler.utils import WorkerContractorPool
from sampo.schemas.resources import Worker
from sampo.schemas.time import Time


class AverageReqResourceOptimizer(ResourceOptimizer):
    """
    Class that implements optimization the number of resources by counting average resource requirements.
    """

    def __init__(self, k: float = 2):
        """
        :param k: coefficient of average
        """
        self.k = k

    def optimize_resources(self,
                           worker_pool: WorkerContractorPool,
                           worker_team: list[Worker],
                           optimize_array: np.ndarray,
                           down_border: np.ndarray,
                           up_border: np.ndarray,
                           get_finish_time: Callable[[list[Worker]], Time]):
        """
        The resource optimization module, that counts average resource requirements.

        :param worker_pool: global resources pool
        :param worker_team: worker team to optimize
        :param optimize_array: a boolean array that says what positions should be optimized
        :param down_border: down border of optimization
        :param up_border: up border of optimization
        :param get_finish_time: optimization function that should give execution time based on worker team
        """

        if optimize_array:
            for worker, up, down, optimize in zip(worker_team, up_border, down_border, optimize_array):
                if optimize:
                    worker.count = max(1, down) + int((up - down) / self.k)
        else:
            # TODO Remove max()
            for worker, up, down in zip(worker_team, up_border, down_border):
                worker.count = max(1, down) + int((up - down) / self.k)



from abc import ABC, abstractmethod
from random import Random
from typing import TypeVar

# import sampo.scheduler

from sampo.api.genetic_api import ChromosomeType, FitnessFunction, Individual, ScheduleGenerationScheme
from sampo.schemas import WorkGraph, Contractor, LandscapeConfiguration, Schedule, GraphNode, Time, WorkTimeEstimator
from sampo.schemas.schedule_spec import ScheduleSpec

T = TypeVar('T')
R = TypeVar('R')


class ComputationalBackend(ABC):
    def __init__(self):
        # scheduler parameters
        self._wg = None
        self._contractors = None
        self._landscape = None
        self._spec = None
        self._rand = Random()
        self._work_estimator = None

        # additional genetic parameters
        self._toolbox = None
        self._selection_size = None
        self._mutate_order = None
        self._mutate_resources = None
        self._mutate_zones = None
        self._deadline = None
        self._weights = None
        self._init_schedules = None
        self._assigned_parent_time = None
        self._fitness_weights = None
        self._sgs_type = None
        self._only_lft_initialization = None
        self._is_multiobjective = None

    @abstractmethod
    def cache_scheduler_info(self,
                             wg: WorkGraph,
                             contractors: list[Contractor],
                             landscape: LandscapeConfiguration,
                             spec: ScheduleSpec,
                             rand: Random | None = None,
                             work_estimator: WorkTimeEstimator | None = None):
        ...

    @abstractmethod
    def cache_genetic_info(self,
                           population_size: int,
                           mutate_order: float,
                           mutate_resources: float,
                           mutate_zones: float,
                           deadline: Time | None,
                           weights: list[int] | None,
                           init_schedules: dict[str, tuple[Schedule, list[GraphNode] | None, ScheduleSpec, float]],
                           assigned_parent_time: Time,
                           fitness_weights: tuple[int | float, ...],
                           sgs_type: ScheduleGenerationScheme,
                           only_lft_initialization: bool,
                           is_multiobjective: bool):
        ...

    @abstractmethod
    def compute_chromosomes(self,
                            fitness: FitnessFunction,
                            chromosomes: list[ChromosomeType]) -> list[float]:
        ...

    @abstractmethod
    def generate_first_population(self, size_population: int) -> list[Individual]:
        ...



from sampo.schemas.interval import IntervalUniform

ONE_SECTION_PIPE = IntervalUniform(0.5, 3)
DIST_TO_PARENT = IntervalUniform(5, 15)
DIST_BETWEEN_BOREHOLES = IntervalUniform(0.05, 0.3)
BRANCHING_PROBABILITY = 0.15


# minimal few pack, optimal for parallel test 20-40
# can be divided into few collaborators? for example 15:15:10 for 40x pack
WORKER_PROPORTIONS = {
    'driver': 6,
    'fitter': 6,
    'handyman': 8,
    'electrician': 2,
    'manager': 3,
    'engineer': 2
}

EQUIPMENTS_PROPORTIONS = {}

MIN_GRAPH_COUNTS = {
    'light_masts': IntervalUniform(1, 3),
    'borehole': IntervalUniform(4, 12),
    'pipe_nodes': IntervalUniform(1, 3),
    'pipe_net': IntervalUniform(2, 8)
}

GRAPH_COUNTS = {
    'light_masts': IntervalUniform(2, 4),
    'borehole': IntervalUniform(8, 28),
    'pipe_nodes': IntervalUniform(1, 5),
    'pipe_net': IntervalUniform(6, 20)
}

ADDITION_CLUSTER_PROBABILITY = 0.25
MAX_BOREHOLES_PER_BLOCK = 14



import json
import os
import pydoc
from abc import ABC, abstractmethod
from itertools import chain
from typing import Generic, TypeVar, Union

import numpy as np
import pandas as pd
from typing_extensions import deprecated

from sampo.utilities.serializers import CUSTOM_FIELD_SERIALIZER, CUSTOM_FIELD_DESERIALIZER, CUSTOM_TYPE_SERIALIZER, \
    CUSTOM_TYPE_DESERIALIZER, default_ndarray_serializer, default_dataframe_serializer, default_ndarray_deserializer, \
    default_dataframe_deserializer, default_np_int_deserializer, default_np_int_serializer, \
    default_np_long_serializer, default_np_long_deserializer

# define type of result serialization
T = TypeVar('T', str, dict, list, tuple, str, bool, None)

# define type of data structure (result of deserialization)
S = TypeVar('S', bound='Serializable')

# define type of deserialized data structure as string
SS = TypeVar('SS', bound='StrSerializable')

# define type of deserialized data structure as JSON
JS = TypeVar('JS', bound='JSONSerializable')

# define type of deserialized data structure as class
AJS = TypeVar('AJS', bound='AutoJSONSerializable')

TYPE_HINTS = '_serializable_type_hints'


# TODO: Implement PartialSerializable, which can't completely deserialize itself, and just returns rich info to parent

class Serializable(ABC, Generic[T, S]):
    """
    Parent class for (de-)serialization different data structures.

    :param ABC: helper class to create custom abstract classes
    :param Generic[T, S]: base class to make Serializable as universal class, using user's types T, S
    """

    @property
    @abstractmethod
    def serializer_extension(self) -> str:
        return NotImplemented

    @abstractmethod
    def _serialize(self) -> T:
        """
        Converts all the meaningful information from this instance to a generic representation.

        :return: A generic representation
        """
        ...

    @classmethod
    @abstractmethod
    def _deserialize(cls, representation: T) -> S:
        """
        Creates class instance from a representation.
        :param representation: Representation produced by _serialize method
        :return: New class instance
        """
        ...

    @classmethod
    @abstractmethod
    def load(cls, folder_path: str, file_name: str) -> S:
        """
        Factory method that produces a python object from the serialized version of it.
        :param folder_path: Path to the folder, where the serialized file is saved
        :param file_name: File name without extension
        (the file extension should match with the one returned by serializer_extension method)
        :return: The constructed python object
        """
        ...

    def dump(self, folder_path: str, file_name: str):
        """
        Serializes the object and saves it to file.
        :param folder_path: Path to the folder where the serialized file should be saved
        :param file_name: Name of the file without extension
        (the appended extension could be explored via serializer_extension method)
        :return None
        """
        ...

    @classmethod
    def get_full_file_name(cls, folder_path: str, file_name: str):
        """
        Combines path to folder, file name and extension to get full file name
        :param folder_path: Path to folder
        :param file_name: File name without extension
        :return: Full file path, name and extension
        """
        return os.path.join(folder_path, f'{file_name}.{cls.serializer_extension}')


class StrSerializable(Serializable[str, SS], ABC, Generic[SS]):
    """
    Parent class for serialization of classes, which can be converted to String representation or given from String
    representation

    :param Serializable[str, SS]:
    :param ABC: helper class to create custom abstract classes
    :param Generic[SS]: base class to make StrSerializable as universal class,
    using user's types SS and it's descendants
    """

    serializer_extension: str = 'dat'

    @abstractmethod
    def _serialize(self) -> str:
        """
        Converts object to str representation
        :return: str representation of the object
        """
        ...

    @classmethod
    @abstractmethod
    def _deserialize(cls, str_representation: str) -> SS:
        """
        Creates class instance from a str representation
        :param str_representation: Representation produced by _serialize method
        :return: New class instance
        """
        ...

    @classmethod
    def load(cls, folder_path: str, file_name: str) -> SS:
        """
        Factory method that produces a python object from the serialized version of it
        :param folder_path: Path to the folder, where the serialized file is saved
        :param file_name: File name without extension
        (the file extension should match with the one returned by serializer_extension method)
        :return: The constructed python object
        """
        full_file_name = cls.get_full_file_name(folder_path, file_name)
        with open(full_file_name, 'r', encoding='utf-8') as read_file:
            str_representation = read_file.read()
        return cls._deserialize(str_representation)

    def dump(self, folder_path: str, file_name: str) -> None:
        """
        Serializes object and saves it to file
        :param folder_path: Path to the folder where the serialized file should be saved
        :param file_name: Name of the file without extension
        (the appended extension could be explored via serializer_extension method)
        :return None
        """
        full_file_name = self.get_full_file_name(folder_path, file_name)
        serialized_str = self._serialize()
        with open(full_file_name, 'w', encoding='utf-8') as write_file:
            write_file.write(serialized_str)


class JSONSerializable(Serializable[dict[str,
                                         Union[
                                             dict,
                                             Union[list, tuple],
                                             str,
                                             Union[int, float],
                                             bool,
                                             None]],
                                    JS], ABC, Generic[JS]):
    serializer_extension: str = 'json'
    """
    Parent class for serialization of classes, which can convert object to JSON format (as serialization result) and get
    object from JSON format
    :param Serialize: structure of JSON input variable for descendants of the class
    :param ABC: helper class to create custom abstract classes
    :param Generic[JS]: base class to make JSONSerializable as universal class, 
    using user's types JS and it's descendants
    """

    @abstractmethod
    def _serialize(self) -> T:
        """
        Converts all the meaningful information from this instance to a generic representation
        :return: A generic representation
        """
        ...

    @classmethod
    @abstractmethod
    def _deserialize(cls, representation: T) -> JS:
        """
        Creates class instance from a representation
        :param representation: Representation produced by _serialize method
        :return: New class instance
        """
        ...

    @classmethod
    def loadd(cls, dict_representation: dict) -> JS:
        """
        Factory method that produces a python object from the serialized version of it
        :param dict_representation: the dictionary repr of the object
        :return: The constructed python object
        """
        return cls._deserialize(dict_representation)

    @classmethod
    def loads(cls, str_representation: str) -> JS:
        """
        Factory method that produces a python object from the serialized version of it
        :param str_representation: the string repr of the object
        :return: The constructed python object
        """
        return cls.loadd(json.loads(str_representation))

    @classmethod
    @deprecated('Use loadf. Will be removed in version 0.3')
    def load(cls, folder_path: str, file_name: str) -> JS:
        """
        Factory method that produces a python object from the serialized version of it
        :param folder_path: Path to the folder, where the serialized file is saved
        :param file_name: File name without extension
        (the file extension should match with the one returned by serializer_extension method)
        :return: The constructed python object
        """
        return cls.loadf(folder_path, file_name)

    @classmethod
    def loadf(cls, folder_path: str, file_name: str) -> JS:
        """
        Factory method that produces a python object from the serialized version of it
        :param folder_path: Path to the folder, where the serialized file is saved
        :param file_name: File name without extension
        (the file extension should match with the one returned by serializer_extension method)
        :return: The constructed python object
        """
        full_file_name = cls.get_full_file_name(folder_path, file_name)
        with open(full_file_name, 'r', encoding='utf-8') as read_file:
            dict_representation = json.load(read_file)
            return cls.loadd(dict_representation)

    def dumpd(self) -> dict:
        """
        Serializes the object and returns it as a dict.
        :return serialized object
        """
        return self._serialize()

    def dumps(self) -> str:
        """
        Serializes the object and returns it as a string.
        :return serialized object
        """
        return json.dumps(self.dumpd())

    def dump(self, folder_path: str, file_name: str):
        """
        Serializes the object and saves it to file.
        :param folder_path: Path to the folder where the serialized file should be saved
        :param file_name: Name of the file without extension
        (the appended extension could be explored via serializer_extension method)
        :return None
        """
        full_file_name = self.get_full_file_name(folder_path, file_name)
        serialized_str = self.dumps()
        with open(full_file_name, 'w', encoding='utf-8') as write_file:
            write_file.write(serialized_str)


# TODO: Implement automotive Enum serialization/deserialization
# TODO: Implement auto serialization of typed collection
class AutoJSONSerializable(JSONSerializable[AJS], ABC):
    """
    Parent class for serialization of classes, which can be automatically converted to dict with Serializable properties
    and custom (de-)serializers, marked with custom_serializer and custom_deserializer decorators.
    :param JSONSerializable[AJS]:
    :param ABC: helper class to create custom abstract classes
    """
    serializer_extension: str = 'json'

    _default_serializers_deserializers = {
        'serializers':
            {
                str(np.ndarray): default_ndarray_serializer,
                str(pd.DataFrame): default_dataframe_serializer,
                str(np.int32): default_np_int_serializer,
                str(np.int64): default_np_long_serializer,
            },
        'deserializer':
            {
                str(np.ndarray): default_ndarray_deserializer,
                str(pd.DataFrame): default_dataframe_deserializer,
                str(np.int32): default_np_int_deserializer,
                str(np.int64): default_np_long_deserializer,
            }
    }

    @classmethod
    @property
    def _default_serializers(cls) -> dict:
        """
        :param cls: current class (predecessors of parent class AutoJSONSerializable)
        :return: dict: dictionary of serialization methods
        """
        return cls._default_serializers_deserializers['serializers']

    @classmethod
    @property
    def _default_deserializers(cls):
        """
        :param cls: current class (predecessors of parent class AutoJSONSerializable)
        :return: dictionary of deserialization methods
        """
        return cls._default_serializers_deserializers['deserializer']

    @classmethod
    @property
    def ignored_fields(cls) -> list:
        """
        Return list of fields, which not be included to JSON representation (needed for client interface
        and Draw Schedule system)
        :return: list of fields
        """
        return []

    def _serialize(self) -> dict[str,
                                 Union[
                                     dict,
                                     Union[list, tuple],
                                     str,
                                     Union[int, float],
                                     bool,
                                     None]]:
        """
        Converts all the meaningful information from this instance to dict representation
        with values of eligible types only to be converted to JSON data structures:
            - list or tuple (to array)
            - str (to string)
            - int or float (to number)
            - True or False (to true or false)
            - None (to null)
            - dict with values of the listed types (to object)
        :return: dict representation of the object
        """
        simple_types = dict, list, tuple, str, int, float, bool, type(None)
        custom_field_serializers = dict(chain(*[[(field, attr) for field in getattr(attr, CUSTOM_FIELD_SERIALIZER)]
                                                for attr in [getattr(self, name) for name in dir(self)]
                                                if hasattr(attr, CUSTOM_FIELD_SERIALIZER)]))
        custom_type_serializers = dict(chain(*[[(__type, attr) for __type in getattr(attr, CUSTOM_TYPE_SERIALIZER)]
                                               for attr in [getattr(self, name) for name in dir(self)]
                                               if hasattr(attr, CUSTOM_TYPE_SERIALIZER)]))
        type_hints = {}

        def serialize_field(name, value):
            """
            Try to serialize value to and put it into resulting dict with key 'name'
            :param name: key of value in resulting dict (JSON)
            :param value:
            :return: serialized value
            """
            if name in custom_field_serializers:
                return custom_field_serializers[name](value)

            t = type(value)
            str_t = str(t)
            name_t = t.__name__
            if t in custom_type_serializers:
                type_hints[name] = str_t
                return custom_type_serializers[t](value)
            if str_t in custom_type_serializers:
                type_hints[name] = str_t
                return custom_type_serializers[str_t](value)
            if name_t in custom_type_serializers:
                type_hints[name] = name_t
                return custom_type_serializers[name_t](value)
            if str_t in self._default_serializers:
                type_hints[name] = str_t
                return self._default_serializers[str_t](value)
            if issubclass(t, Serializable):
                type_hints[name] = pydoc.classname(t, '')
                return value._serialize()
            if t in simple_types:
                return value
            raise ValueError(f'Cannot serialize field {name}.'
                             f' Provide custom serializer for it (via decorator @custom_serializer)'
                             f' or make inherit {type(value)} from Serializable.')

        return dict({TYPE_HINTS: type_hints},
                    **{k: serialize_field(k, v) for k, v in self.__dict__.items() if k not in self.ignored_fields})

    @classmethod
    def _deserialize(cls, dict_representation: dict) -> AJS:
        """
        Creates class instance from a dict representation

        :param dict_representation: Representation produced by _serialize method
        :return: New class instance
        """
        custom_field_deserializers = dict(chain(*[[(field, attr) for field in getattr(attr, CUSTOM_FIELD_DESERIALIZER)]
                                                for attr in [getattr(cls, name) for name in dir(cls)]
                                                if hasattr(attr, CUSTOM_FIELD_DESERIALIZER)]))
        custom_type_deserializers = dict(chain(*[[(str(__type), attr) for __type in getattr(attr,
                                                                                          CUSTOM_TYPE_DESERIALIZER)]
                                               for attr in [getattr(cls, name) for name in dir(cls)]
                                               if hasattr(attr, CUSTOM_TYPE_DESERIALIZER)]))
        type_hints = {}
        if TYPE_HINTS in dict_representation:
            type_hints = dict_representation[TYPE_HINTS]
            del dict_representation[TYPE_HINTS]

        def deserialize_field(name, value):
            """
            Transform current value from JSON to element of resulting dict, having 'name' as key
            :param name: key in dict
            :param value:
            :return: deserialize value
            """
            if name in custom_field_deserializers:
                return custom_field_deserializers[name](value)
            if name in type_hints:
                __type = type_hints[name]
                if __type in custom_type_deserializers:
                    return custom_type_deserializers[__type](value)
                if __type in cls._default_deserializers:
                    return cls._default_deserializers[__type](value)
                c = pydoc.locate(type_hints[name])
                return c._deserialize(value)
            return value

        fields = dict({f: None for f in cls.ignored_fields},
                      **{k: deserialize_field(k, v)
                         for k, v in dict_representation.items()
                         if k not in cls.ignored_fields})

        self = cls.__new__(cls)
        try:
            self.__dict__ = fields
        except:
            for k, v in fields.items():
                object.__setattr__(self, k, v)
        return self



import math
import random
import uuid
from collections import defaultdict

from sampo.schemas import Material, WorkGraph
from sampo.schemas.landscape import ResourceHolder, Vehicle, LandscapeConfiguration
from sampo.schemas.landscape_graph import LandGraphNode, ResourceStorageUnit, LandGraph


def setup_landscape(platforms_info: dict[str, dict[str, int]],
                    warehouses_info: dict[str, list[dict[str, int], list[tuple[str, dict[str, int]]]]],
                    roads_info: dict[str, list[tuple[str, float, int]]]) -> LandscapeConfiguration:
    """
    Build landscape configuration based on the provided information with structure as below.
    Attributes:
           Platform_info structure:
                {platform_name:
                    {material_name: material_count}
                }
           Warehouse_info structure:
                {holder_name:
                    [
                        {material_name: material_count},
                        [(vehicle_name, {vehicle_material_name: vehicle_material_count})]
                    ]
                }
           Roads_info structure:
                {platform_name:
                    [(neighbour_name, road_length, road_workload)]
                }
    :return: landscape configuration
    """
    name2platform: dict[str, LandGraphNode] = {}
    holders: list[ResourceHolder] = []
    for platform, platform_info in platforms_info.items():
        node = LandGraphNode(str(uuid.uuid4()), platform, ResourceStorageUnit(
                {name: count for name, count in platform_info.items()}
            ))
        name2platform[platform] = node

    for holder_name, holder_info in warehouses_info.items():
        materials = holder_info[0]
        vehicles = holder_info[1]
        holder_node = LandGraphNode(
            str(uuid.uuid4()), holder_name, ResourceStorageUnit(
                {name: count for name, count in materials.items()}
            ))
        name2platform[holder_name] = holder_node
        holders.append(ResourceHolder(
            str(uuid.uuid4()), holder_name,
            [
                Vehicle(str(uuid.uuid4()), name, [
                    Material(str(uuid.uuid4()), mat_name, mat_count)
                    for mat_name, mat_count in vehicle_mat_info.items()
                    ])
                for name, vehicle_mat_info in vehicles
            ],
            holder_node
        ))

    for from_node, adj_list in roads_info.items():
        name2platform[from_node].add_neighbours([(name2platform[node], length, workload)
                                                 for node, length, workload in adj_list])

    platforms: list[LandGraphNode] = list(name2platform.values())

    return LandscapeConfiguration(
        holders=holders,
        lg=LandGraph(nodes=platforms)
    )


def get_landscape_by_wg(wg: WorkGraph, rnd: random.Random) -> LandscapeConfiguration:
    nodes = wg.nodes
    max_materials = defaultdict(int)

    for node in nodes:
        for mat in node.work_unit.need_materials():
            if mat.name not in max_materials:
                max_materials[mat.name] = mat.count
            else:
                max_materials[mat.name] = max(max_materials[mat.name], mat.count)

    platforms_number = math.ceil(math.log(wg.vertex_count))
    platforms = []
    materials_name = list(max_materials.keys())

    for i in range(platforms_number):
        platforms.append(LandGraphNode(str(uuid.uuid4()), f'platform{i}',
                                       ResourceStorageUnit(
                                           {
                                               name: rnd.randint(max(max_materials[name], 1),
                                                                 2 * max(max_materials[name], 1))
                                               # name: max(max_materials[name], 1)
                                               for name in materials_name
                                           }
                                       )))

    for i, platform in enumerate(platforms):
        if i == platforms_number - 1:
            continue
        neighbour_platforms = rnd.choices(platforms[i + 1:], k=rnd.randint(1, math.ceil(len(platforms[i + 1:]) / 3)))

        neighbour_platforms_tmp = neighbour_platforms.copy()
        for neighbour in neighbour_platforms:
            if neighbour in platform.neighbours:
                neighbour_platforms_tmp.remove(neighbour)
        neighbour_platforms = neighbour_platforms_tmp

        # neighbour_edges = [(neighbour, rnd.uniform(1.0, 10.0), rnd.randint(wg.vertex_count, wg.vertex_count * 2))
        #                    for neighbour in neighbour_platforms]
        neighbour_edges = [(neighbour,  20.0, wg.vertex_count * 2)
                           for neighbour in neighbour_platforms]
        platform.add_neighbours(neighbour_edges)

    inseparable_heads = [node for node in nodes if not node.is_inseparable_son()]

    platforms_tmp = ((len(inseparable_heads) // platforms_number) * platforms +
                     platforms[:len(inseparable_heads) % platforms_number])
    rnd.shuffle(platforms_tmp)

    for node, platform in zip(inseparable_heads, platforms_tmp):
        if not node.work_unit.is_service_unit:
            for ins_child in node.get_inseparable_chain_with_self():
                platform.add_works(ins_child)

    holders_number = math.ceil(math.sqrt(math.log(wg.vertex_count)))
    holders_node = []
    holders = []

    sample_materials_for_holders = materials_name * holders_number
    # random.shuffle(sample_materials_for_holders)
    materials_number = len(materials_name)

    for i in range(holders_number):
        if not max_materials:
            materials_name_for_holder = []
        else:
            materials_name_for_holder = sample_materials_for_holders[i * materials_number: (i + 1) * materials_number]
        holders_node.append(LandGraphNode(str(uuid.uuid4()), f'holder{i}',
                                          ResourceStorageUnit(
                                              {
                                                  name: max(max_materials[name], 1) * wg.vertex_count
                                                  for name in materials_name_for_holder
                                              }
                                          )))
        neighbour_platforms = rnd.choices(holders_node[:-1] + platforms, k=rnd.randint(1, len(holders_node[:-1] + platforms)))

        neighbour_platforms_tmp = neighbour_platforms.copy()
        for neighbour in neighbour_platforms:
            if neighbour in holders_node[-1].neighbours:
                neighbour_platforms_tmp.remove(neighbour)
        neighbour_platforms = neighbour_platforms_tmp

        # neighbour_edges = [(neighbour, rnd.uniform(1.0, 10.0), rnd.randint(wg.vertex_count, wg.vertex_count * 2))
        #                    for neighbour in neighbour_platforms]
        neighbour_edges = [(neighbour, 20.0, wg.vertex_count * 2)
                           for neighbour in neighbour_platforms]
        holders_node[-1].add_neighbours(neighbour_edges)

        vehicles_number = 20
        holders.append(ResourceHolder(str(uuid.uuid4()), holders_node[-1].name,
                                      vehicles=[
                                          Vehicle(str(uuid.uuid4()), f'vehicle{j}',
                                                  [Material(name, name, count)
                                                   for name, count in max_materials.items()])
                                          for j in range(vehicles_number)
                                      ], node=holders_node[-1]))

    lg = LandGraph(nodes=platforms + holders_node)
    return LandscapeConfiguration(holders, lg)



from typing import Callable

import numpy as np

from sampo.scheduler.resource.base import ResourceOptimizer
from sampo.scheduler.resource.coordinate_descent import CoordinateDescentResourceOptimizer
from sampo.scheduler.utils import WorkerContractorPool
from sampo.schemas.resources import Worker
from sampo.schemas.time import Time
from sampo.utilities.base_opt import dichotomy_int


class FullScanResourceOptimizer(ResourceOptimizer):
    """
    Class that implements optimization the number of resources by the smart search method.
    """

    _coordinate_descent_optimizer = CoordinateDescentResourceOptimizer(dichotomy_int)

    def optimize_resources(self,
                           worker_pool: WorkerContractorPool,
                           worker_team: list[Worker],
                           optimize_array: np.ndarray,
                           down_border: np.ndarray,
                           up_border: np.ndarray,
                           get_finish_time: Callable[[list[Worker]], Time]):
        """
        The resource optimization module, that search optimal number of resources by the smart search method.

        :param worker_pool: global resources pool
        :param worker_team: worker team to optimize
        :param optimize_array: a boolean array that says what positions should be optimized
        :param down_border: down border of optimization
        :param up_border: up border of optimization
        :param get_finish_time: optimization function that should give execution time based on worker team
        """

        # TODO Handle optimize_array
        def fitness(worker_count: np.ndarray):
            """
            Function that counts finish time with a current set of resources

            :param worker_count:
            :return: finish time with a current set of resources
            """
            for ind, worker in enumerate(worker_team):
                worker.count = worker_count[ind]
            return get_finish_time(worker_team)

        right_fit = fitness(up_border)
        left_fit = fitness(down_border)

        mid = up_border

        # Find the optimal point on the search area diagonal
        while (mid > down_border).any():
            mid = (up_border + down_border) // 2
            next_fit = fitness(mid)
            if next_fit < right_fit:
                down_border = mid
                left_fit = next_fit
            elif next_fit > left_fit:
                up_border = mid
                right_fit = next_fit
            else:
                break

        # Use coordinate decent to search an optimal set of resources with new down_border
        self._coordinate_descent_optimizer.optimize_resources(worker_pool,
                                                              worker_team,
                                                              optimize_array,
                                                              mid,
                                                              up_border,
                                                              get_finish_time)



from abc import ABC, abstractmethod
from enum import Enum
from operator import attrgetter
from random import Random
from typing import Optional, Type

import numpy.random
import math

from sampo.schemas import Interval, IntervalGaussian
from sampo.schemas.requirements import WorkerReq
from sampo.schemas.resources import Worker
from sampo.schemas.resources import WorkerProductivityMode
from sampo.schemas.time import Time
from sampo.schemas.works import WorkUnit
from sampo.utilities.collections_util import build_index


class WorkEstimationMode(Enum):
    Pessimistic = -1,
    Realistic = 0,
    Optimistic = 1


class WorkTimeEstimator(ABC):
    """
    Implementation of time estimator of work with a given set of resources.
    """

    @abstractmethod
    def set_estimation_mode(self, use_idle: bool = True, mode: WorkEstimationMode = WorkEstimationMode.Realistic):
        ...

    @abstractmethod
    def set_productivity_mode(self, mode: WorkerProductivityMode = WorkerProductivityMode.Static):
        ...

    @abstractmethod
    def find_work_resources(self, work_name: str, work_volume: float, resource_name: list[str] | None = None) \
            -> list[WorkerReq]:
        ...

    @abstractmethod
    def estimate_time(self, work_unit: WorkUnit, worker_list: list[Worker]):
        ...

    @abstractmethod
    def get_recreate_info(self) -> tuple[Type, tuple]:
        ...


class DefaultWorkEstimator(WorkTimeEstimator):

    def __init__(self,
                 rand: Random = Random()):
        self._use_idle = True
        self._estimation_mode = WorkEstimationMode.Realistic
        self.rand = rand
        self._productivity_mode = WorkerProductivityMode.Static
        self._productivity = {worker: {'__ALL__': IntervalGaussian(1, 0.2, 1, 0)}
                              for worker in ['driver', 'fitter', 'manager', 'handyman', 'electrician', 'engineer']}

    def find_work_resources(self, work_name: str, work_volume: float, resource_name: list[str] | None = None) \
            -> list[WorkerReq]:
        if resource_name is None:
            resource_name = ['driver', 'fitter', 'manager', 'handyman', 'electrician', 'engineer']
        dist = numpy.random.poisson(work_volume * 3, len(resource_name))
        return [WorkerReq(kind=name,
                          volume=Time(math.ceil(work_volume * numpy.random.poisson(work_volume ** 0.5, 1)[0])),
                          min_count=int(dist[i]),
                          max_count=int(dist[i] * 2))
                for i, name in enumerate(resource_name)]

    def set_estimation_mode(self, use_idle: bool = True, mode: WorkEstimationMode = WorkEstimationMode.Realistic):
        self._use_idle = use_idle
        self._estimation_mode = mode

    def set_productivity_mode(self, mode: WorkerProductivityMode = WorkerProductivityMode.Static):
        self._productivity_mode = mode

    def estimate_time(self, work_unit: WorkUnit, worker_list: list[Worker]) -> Time:
        if not worker_list:
            return Time(0)

        times = [Time(0)]  # if there are no requirements for the work, it is done instantly
        name2worker = build_index(worker_list, attrgetter('name'))

        for req in work_unit.worker_reqs:
            if req.min_count == 0:
                continue
            name = req.kind
            worker = name2worker.get(name, None)
            worker_count = 0 if worker is None else worker.count
            if worker_count < req.min_count:
                return Time.inf()
            productivity = self._get_productivity(worker, req.max_count)
            if productivity == 0:
                return Time.inf()
            times.append(Time(math.ceil(req.volume / productivity)))
        return max(max(times), Time(0))

    def _get_productivity(self, worker: Worker, max_count_workers: int) -> float:
        """
        Calculate the productivity of the Worker
        It has 2 modes: stochastic and non-stochastic, depending on the value of rand
        :param worker: the worker
        :param max_count_workers: maximum workers count according to worker reqs
        :return:
        """
        worker_productivities = self._productivity[worker.name]
        productivity_interval = worker_productivities.get(worker.contractor_id, worker_productivities['__ALL__'])
        productivity = productivity_interval.mean \
            if self._productivity_mode is WorkerProductivityMode.Static \
            else productivity_interval.rand_float(self.rand)
        return productivity * worker.count * communication_coefficient(worker.count, max_count_workers)

    def set_worker_productivity(self, productivity: Interval, name: str, contractor: str | None = None):
        if contractor is None:
            self._productivity[name]['__ALL__'] = productivity
            return
        self._productivity[name][contractor] = productivity

    def get_recreate_info(self) -> tuple[Type, tuple]:
        return DefaultWorkEstimator, ()


def communication_coefficient(groups_count: int, max_groups: int) -> float:
    n = groups_count
    m = max_groups
    return 1.0 / (6 * m ** 2) * (-2 * n ** 3 + 3 * n ** 2 + (6 * m ** 2 - 1) * n)



from enum import Enum, auto
from sampo.schemas import GraphNode

StageType = tuple[GraphNode, dict[str, GraphNode]]


class SyntheticGraphType(Enum):
    """
    Describe available types of synthetic graph

    * PARALLEL - work graph dominated by parallel works
    * SEQUENTIAL - work graph dominated by sequential works
    * GENERAL - work graph, including sequential and parallel works, it is similar to the real work graphs
    """
    GENERAL = auto()
    PARALLEL = auto()
    SEQUENTIAL = auto()


from sampo.generator.pipeline.cluster import get_cluster_works
from sampo.generator.pipeline.extension import extend_names, extend_resources
from sampo.generator.pipeline.project import get_small_graph, get_graph, get_cluster_works



from abc import ABC, abstractmethod
from enum import Enum
from typing import Callable

import numpy as np

from sampo.scheduler.resource.base import ResourceOptimizer
from sampo.scheduler.resource.coordinate_descent import CoordinateDescentResourceOptimizer
from sampo.scheduler.timeline.base import Timeline
from sampo.schemas.contractor import Contractor
from sampo.schemas.graph import WorkGraph, GraphNode
from sampo.schemas.landscape import LandscapeConfiguration
from sampo.schemas.resources import Worker
from sampo.schemas.schedule import Schedule
from sampo.schemas.schedule_spec import ScheduleSpec, WorkSpec
from sampo.schemas.time import Time
from sampo.schemas.time_estimator import WorkTimeEstimator, DefaultWorkEstimator
from sampo.schemas.works import WorkUnit
from sampo.utilities.base_opt import dichotomy_int


class SchedulerType(Enum):
    Genetic = 'genetic'
    Topological = 'topological'
    HEFTAddEnd = 'heft_add_end'
    HEFTAddBetween = 'heft_add_between'
    LFT = 'LFT'


class Scheduler(ABC):
    """
    Base class that implements the logic of the planning process.
    """
    scheduler_type: SchedulerType
    resource_optimizer: ResourceOptimizer

    def __init__(self,
                 scheduler_type: SchedulerType,
                 resource_optimizer: ResourceOptimizer = CoordinateDescentResourceOptimizer(dichotomy_int),
                 work_estimator: WorkTimeEstimator = DefaultWorkEstimator()):
        self.scheduler_type = scheduler_type
        self.resource_optimizer = resource_optimizer
        self.work_estimator = work_estimator

    def __str__(self) -> str:
        return str(self.scheduler_type.name)

    def schedule(self,
                 wg: WorkGraph,
                 contractors: list[Contractor],
                 spec: ScheduleSpec = ScheduleSpec(),
                 validate: bool = False,
                 start_time: Time = Time(0),
                 timeline: Timeline | None = None,
                 landscape: LandscapeConfiguration = LandscapeConfiguration()) \
            -> list[Schedule]:
        """
        Implementation of a scheduling process. 'schedule' version returns only Schedule.

        :return: Schedule
        """
        if wg is None or len(wg.nodes) == 0:
            raise ValueError('None or empty WorkGraph')
        if contractors is None or len(contractors) == 0:
            raise ValueError('None or empty contractor list')
        schedules = self.schedule_with_cache(wg, contractors, spec, validate, start_time, timeline, landscape)
        schedules = [schedule[0] for schedule in schedules]
        # print(f'Schedule exec time: {schedule.execution_time} days')
        return schedules

    @abstractmethod
    def schedule_with_cache(self,
                            wg: WorkGraph,
                            contractors: list[Contractor],
                            spec: ScheduleSpec = ScheduleSpec(),
                            validate: bool = False,
                            assigned_parent_time: Time = Time(0),
                            timeline: Timeline | None = None,
                            landscape: LandscapeConfiguration = LandscapeConfiguration()) \
            -> list[tuple[Schedule, Time, Timeline, list[GraphNode]]]:
        """
        Extended version of 'schedule' method. Returns much inner info
        about a scheduling process, not only Schedule.

        :return: resulting schedule, finish time,
                 resulting timeline used for scheduling
                 and node_order used for scheduling
        """
        ...

    @staticmethod
    def optimize_resources_using_spec(work_unit: WorkUnit, worker_team: list[Worker], work_spec: WorkSpec,
                                      optimize_lambda: Callable[[np.ndarray], None] = lambda _: None):
        """
        Applies worker team spec to an optimization process.
        Can use arbitrary heuristics to increase spec handling efficiency.

        :param work_unit: current work unit
        :param worker_team: current worker team from chosen contractor
        :param work_spec: spec for given work unit
        :param optimize_lambda: optimization func that should hold optimization
            data in its closure and run optimization process when receives `optimize_array`.
            Passing None or default value means this function should only apply spec.
        """
        if len(work_spec.assigned_workers) == len(work_unit.worker_reqs):
            # all resources passed in spec, skipping optimize_resources step
            for worker in worker_team:
                worker.count = work_spec.assigned_workers[worker.name]
        else:
            # create optimize array to save optimizing time
            # this array should contain True if position should be optimized or False if shouldn't
            optimize_array = None
            if work_spec.assigned_workers:
                optimize_array = []
                for worker in worker_team:
                    spec_count = work_spec.assigned_workers.get(worker.name, 0)
                    if spec_count > 0:
                        worker.count = spec_count
                        optimize_array.append(False)
                    else:
                        optimize_array.append(True)

                optimize_array = np.array(optimize_array)

            optimize_lambda(optimize_array)



from dataclasses import dataclass
from enum import Enum
from random import Random
from typing import Optional

from sampo.schemas.identifiable import Identifiable
from sampo.schemas.interval import IntervalGaussian
from sampo.schemas.serializable import AutoJSONSerializable
from sampo.schemas.types import AgentId


class WorkerProductivityMode(Enum):
    Static = 'static',
    Stochastic = 'stochastic'


@dataclass
class Resource(AutoJSONSerializable['Resource'], Identifiable):
    """
    A class summarizing the different resources used in the work: Human resources, equipment, materials, etc.
    """
    id: str
    name: str
    count: int
    contractor_id: Optional[str] = ""

    def __init__(self,
                 id: str,
                 name: str,
                 count: int,
                 contractor_id: Optional[str] = ""):
        self.id = id
        self.name = name
        self._count = count
        self.contractor_id = contractor_id

    @property
    def count(self) -> int:
        return self._count

    @count.setter
    def count(self, value):
        self._count = int(value)

    # TODO: describe the function (description, return type)
    def get_agent_id(self) -> AgentId:
        return self.contractor_id, self.name


@dataclass
class Worker(Resource):
    """
    A class dedicated to human resources

    :param count: the number of people in this resource
    :param contractor_id: Contractor id if resources are added directly to the contractor
    :param productivity: interval from Gaussian or Uniform distribution, that contains possible values of
    productivity of certain worker
    """

    def __init__(self,
                 id: str,
                 name: str,
                 count: int,
                 contractor_id: Optional[str] = "",
                 productivity: Optional[IntervalGaussian] = IntervalGaussian(1, 0, 1, 1),
                 cost_one_unit: Optional[float] = None):
        super(Worker, self).__init__(id, name, int(count), contractor_id)
        self.productivity = productivity if productivity is not None else IntervalGaussian(1, 0, 1, 1)
        self.cost_one_unit = cost_one_unit if cost_one_unit is not None else self.productivity.mean * 10

    ignored_fields = ['productivity']

    def copy(self) -> 'Worker':
        """
        Return copied current object

        :return: object of Worker class
        """
        return Worker(id=self.id,
                      name=self.name,
                      count=self.count,
                      contractor_id=self.contractor_id,
                      productivity=self.productivity)

    def with_count(self, count: int) -> 'Worker':
        """
        Update count field

        :param count: amount of current type of worker
        :return: upgraded current object
        """
        self.count = int(count)
        return self

    def get_cost(self) -> float:
        """Returns cost of this worker entry"""
        return self.cost_one_unit * int(self.count)

    def get_agent_id(self) -> AgentId:
        """
        Return the agent unique identifier in schedule

        :return: contractor's id and name
        """
        return self.contractor_id, self.name

    def get_productivity(self, rand: Random, productivity_mode: WorkerProductivityMode) -> float:
        """
        Return the productivity of the worker team.
        It has 2 mods: stochastic and non-stochastic.

        :param productivity_mode:
        :param rand: parameter for stochastic part
        """
        if productivity_mode is WorkerProductivityMode.Static:
            return self.productivity.mean * self.count
        return self.productivity.rand_float(rand) * self.count

    def __repr__(self) -> str:
        return f'{self.count} {self.name}'

    def __str__(self) -> str:
        return self.__repr__()


@dataclass
class ConstructionObject(Resource):
    pass


@dataclass(init=False)
class EmptySpaceConstructionObject(ConstructionObject):
    id: str = '00000000000000000'
    name: str = 'empty space construction object'


@dataclass
class Equipment(Resource):
    pass


@dataclass
class Material(Resource):

    def __init__(self,
                 id: str,
                 name: str,
                 count: int,
                 cost_one_unit: Optional[float] = 1):
        super(Material, self).__init__(id, name, count)
        self.cost_one_unit = cost_one_unit

    # TODO: describe the function (description, return type)
    def copy(self) -> 'Material':
        return Material(id=self.id,
                        name=self.name,
                        count=self.count)

    def with_count(self, count: int) -> 'Material':
        self.count = count
        return self



import random
import time

from deap import tools
from deap.base import Toolbox

from sampo.api.genetic_api import Individual
from sampo.base import SAMPO
from sampo.scheduler.genetic.converter import convert_schedule_to_chromosome, ScheduleGenerationScheme
from sampo.scheduler.genetic.operators import init_toolbox, ChromosomeType, FitnessFunction, TimeFitness
from sampo.scheduler.genetic.utils import prepare_optimized_data_structures
from sampo.scheduler.timeline.base import Timeline
from sampo.schemas.contractor import Contractor
from sampo.schemas.graph import GraphNode, WorkGraph
from sampo.schemas.landscape import LandscapeConfiguration
from sampo.schemas.schedule import ScheduleWorkDict, Schedule
from sampo.schemas.schedule_spec import ScheduleSpec
from sampo.schemas.time import Time
from sampo.schemas.time_estimator import WorkTimeEstimator, DefaultWorkEstimator


def create_toolbox(wg: WorkGraph,
                   contractors: list[Contractor],
                   selection_size: int,
                   mutate_order: float,
                   mutate_resources: float,
                   mutate_zones: float,
                   init_schedules: dict[
                       str, tuple[Schedule, list[GraphNode] | None, ScheduleSpec, float]],
                   rand: random.Random,
                   spec: ScheduleSpec = ScheduleSpec(),
                   fitness_weights: tuple[int | float, ...] = (-1,),
                   work_estimator: WorkTimeEstimator = None,
                   sgs_type: ScheduleGenerationScheme = ScheduleGenerationScheme.Parallel,
                   assigned_parent_time: Time = Time(0),
                   landscape: LandscapeConfiguration = LandscapeConfiguration(),
                   only_lft_initialization: bool = False,
                   is_multiobjective: bool = False,
                   verbose: bool = True) -> Toolbox:
    start = time.time()

    worker_pool, index2node, index2zone, work_id2index, worker_name2index, index2contractor_obj, \
        worker_pool_indices, contractor2index, contractor_borders, node_indices, parents, children, \
        resources_border = prepare_optimized_data_structures(wg, contractors, landscape)

    init_chromosomes: dict[str, tuple[ChromosomeType, float, ScheduleSpec]] = \
        {name: (convert_schedule_to_chromosome(work_id2index, worker_name2index,
                                               contractor2index, contractor_borders, schedule, chromosome_spec,
                                               landscape, order),
                importance, chromosome_spec)
         if schedule is not None else None
         for name, (schedule, order, chromosome_spec, importance) in init_schedules.items()}

    if verbose:
        print(f'Genetic optimizing took {(time.time() - start) * 1000} ms')

    return init_toolbox(wg,
                        contractors,
                        worker_pool,
                        landscape,
                        index2node,
                        work_id2index,
                        worker_name2index,
                        index2contractor_obj,
                        index2zone,
                        init_chromosomes,
                        mutate_order,
                        mutate_resources,
                        mutate_zones,
                        landscape.zone_config.statuses.statuses_available(),
                        selection_size,
                        rand,
                        spec,
                        worker_pool_indices,
                        contractor2index,
                        contractor_borders,
                        node_indices,
                        parents,
                        children,
                        resources_border,
                        assigned_parent_time,
                        fitness_weights,
                        work_estimator,
                        sgs_type,
                        only_lft_initialization,
                        is_multiobjective)


def build_schedules(wg: WorkGraph,
                    contractors: list[Contractor],
                    population_size: int,
                    generation_number: int,
                    mutpb_order: float,
                    mutpb_res: float,
                    mutpb_zones: float,
                    init_schedules: dict[str, tuple[Schedule, list[GraphNode] | None, ScheduleSpec, float]],
                    rand: random.Random,
                    spec: ScheduleSpec,
                    weights: list[int],
                    landscape: LandscapeConfiguration = LandscapeConfiguration(),
                    fitness_object: FitnessFunction = TimeFitness(),
                    fitness_weights: tuple[int | float, ...] = (-1,),
                    work_estimator: WorkTimeEstimator = DefaultWorkEstimator(),
                    sgs_type: ScheduleGenerationScheme = ScheduleGenerationScheme.Parallel,
                    assigned_parent_time: Time = Time(0),
                    timeline: Timeline | None = None,
                    time_border: int | None = None,
                    max_plateau_steps: int | None = None,
                    optimize_resources: bool = False,
                    deadline: Time | None = None,
                    only_lft_initialization: bool = False,
                    is_multiobjective: bool = False) \
        -> list[tuple[ScheduleWorkDict, Time, Timeline, list[GraphNode]]]:
    """
    Genetic algorithm.
    Structure of chromosome:
    [[order of job],
     [[numbers of workers types 1 for each job], [numbers of workers types 2], ... ],
      [[border of workers types 1 for each contractor], [border of workers types 2], ...]
    ]

    Different mate and mutation for order and for workers.
    Generate order of job by prioritization from HEFTs and from Topological.
    Generate resources from min to max.
    Overall initial population is valid.

    :return: schedule
    """
    global_start = start = time.time()

    toolbox = create_toolbox(wg, contractors, population_size,
                             mutpb_order, mutpb_res, mutpb_zones, init_schedules,
                             rand, spec, fitness_weights, work_estimator,
                             sgs_type, assigned_parent_time, landscape,
                             only_lft_initialization, is_multiobjective)

    SAMPO.backend.cache_scheduler_info(wg, contractors, landscape, spec, rand, work_estimator)
    SAMPO.backend.cache_genetic_info(population_size,
                                     mutpb_order, mutpb_res, mutpb_zones,
                                     deadline, weights,
                                     init_schedules, assigned_parent_time, fitness_weights,
                                     sgs_type, only_lft_initialization, is_multiobjective)

    # create population of a given size
    pop = SAMPO.backend.generate_first_population(population_size)

    SAMPO.logger.info(f'Toolbox initialization & first population took {(time.time() - start) * 1000} ms')

    have_deadline = deadline is not None
    fitness_f = fitness_object if not have_deadline else TimeFitness()
    if have_deadline:
        toolbox.register_individual_constructor((-1,))
    evaluation_start = time.time()

    hof = tools.ParetoFront(similar=compare_individuals)

    # map to each individual fitness function
    fitness = SAMPO.backend.compute_chromosomes(fitness_f, pop)

    evaluation_time = time.time() - evaluation_start

    for ind, fit in zip(pop, fitness):
        ind.fitness.values = fit

    hof.update(pop)
    best_fitness = hof[0].fitness.values

    SAMPO.logger.info(f'First population evaluation took {evaluation_time * 1000} ms')

    start = time.time()

    generation = 1
    plateau_steps = 0
    new_generation_number = generation_number if not have_deadline else generation_number // 2
    new_max_plateau_steps = max_plateau_steps if max_plateau_steps is not None else new_generation_number

    while generation <= new_generation_number and plateau_steps < new_max_plateau_steps \
            and (time_border is None or time.time() - global_start < time_border):
        SAMPO.logger.info(f'-- Generation {generation}, population={len(pop)}, best fitness={best_fitness} --')

        rand.shuffle(pop)

        offspring = make_offspring(toolbox, pop, optimize_resources)

        evaluation_start = time.time()

        offspring_fitness = SAMPO.backend.compute_chromosomes(fitness_f, offspring)

        for ind, fit in zip(offspring, offspring_fitness):
            ind.fitness.values = fit

        evaluation_time += time.time() - evaluation_start

        # renewing population
        pop += offspring
        pop = toolbox.select(pop)
        hof.update(pop)

        prev_best_fitness = best_fitness
        best_fitness = hof[0].fitness.values
        plateau_steps = plateau_steps + 1 if best_fitness == prev_best_fitness else 0

        if have_deadline and best_fitness[0] <= deadline:
            if all([ind.fitness.values[0] <= deadline for ind in pop]):
                break

        generation += 1

    # Second stage to optimize resources if deadline is assigned

    if have_deadline:
        fitness_resource_f = fitness_object
        toolbox.register_individual_constructor(fitness_weights)

        SAMPO.logger.info(f'Deadline not reached !!! Deadline {deadline} < best time {best_fitness}')
        # clear best individuals
        hof.clear()

        for ind in pop:
            ind.time = ind.fitness.values[0]

        if best_fitness[0] > deadline:
            SAMPO.logger.info(f'Deadline not reached !!! Deadline {deadline} < best time {best_fitness[0]}')
            pop = [ind for ind in pop if ind.time == best_fitness[0]]
        else:
            optimize_resources = True
            pop = [ind for ind in pop if ind.time <= deadline]

        new_pop = []
        for ind in pop:
            ind_time = ind.time
            new_ind = toolbox.copy_individual(ind)
            new_ind.time = ind_time
            new_pop.append(new_ind)
        del pop
        pop = new_pop

        evaluation_start = time.time()

        fitness = SAMPO.backend.compute_chromosomes(fitness_resource_f, pop)
        for ind, res_fit in zip(pop, fitness):
            ind.fitness.values = res_fit

        evaluation_time += time.time() - evaluation_start

        hof.update(pop)

        if best_fitness[0] <= deadline:
            # Optimizing resources
            plateau_steps = 0
            new_generation_number = generation_number - generation + 1
            new_max_plateau_steps = max_plateau_steps if max_plateau_steps is not None else new_generation_number
            best_fitness = hof[0].fitness.values

            if len(pop) < population_size:
                individuals_to_copy = rand.choices(pop, k=population_size - len(pop))
                copied_individuals = [toolbox.copy_individual(ind) for ind in individuals_to_copy]
                for copied_ind, ind in zip(copied_individuals, individuals_to_copy):
                    copied_ind.fitness.values = ind.fitness.values
                    copied_ind.time = ind.time
                pop += copied_individuals

            while generation <= generation_number and plateau_steps < new_max_plateau_steps \
                    and (time_border is None or time.time() - global_start < time_border):
                SAMPO.logger.info(f'-- Generation {generation}, population={len(pop)}, best peak={best_fitness} --')

                rand.shuffle(pop)

                offspring = make_offspring(toolbox, pop, optimize_resources)

                evaluation_start = time.time()

                fitness = SAMPO.backend.compute_chromosomes(fitness_f, offspring)

                for ind, t in zip(offspring, fitness):
                    ind.time = t[0]

                offspring = [ind for ind in offspring if ind.time <= deadline]

                fitness_res = SAMPO.backend.compute_chromosomes(fitness_resource_f, offspring)

                for ind, res_fit in zip(offspring, fitness_res):
                    ind.fitness.values = res_fit

                evaluation_time += time.time() - evaluation_start

                # renewing population
                pop += offspring
                pop = toolbox.select(pop)
                hof.update(pop)

                prev_best_fitness = best_fitness
                best_fitness = hof[0].fitness.values
                plateau_steps = plateau_steps + 1 if best_fitness == prev_best_fitness else 0

                generation += 1

    SAMPO.logger.info(f'Final fitness: {best_fitness}')
    SAMPO.logger.info(f'Generations processing took {(time.time() - start) * 1000} ms')
    SAMPO.logger.info(f'Full genetic processing took {(time.time() - global_start) * 1000} ms')
    SAMPO.logger.info(f'Evaluation time: {evaluation_time * 1000}')

    best_chromosomes = [chromosome for chromosome in hof]

    best_schedules = [toolbox.chromosome_to_schedule(best_chromosome, landscape=landscape, timeline=timeline)
                      for best_chromosome in best_chromosomes]
    best_schedules = [({node.id: work for node, work in scheduled_works.items()},
                       schedule_start_time, timeline, order_nodes)
                      for scheduled_works, schedule_start_time, timeline, order_nodes in best_schedules]

    return best_schedules


def compare_individuals(first: ChromosomeType, second: ChromosomeType) -> bool:
    return ((first[0] == second[0]).all() and (first[1] == second[1]).all() and (first[2] == second[2]).all()
            or first.fitness == second.fitness)


def make_offspring(toolbox: Toolbox, population: list[ChromosomeType], optimize_resources: bool) \
        -> list[Individual]:
    offspring = []

    for ind1, ind2 in zip(population[::2], population[1::2]):
        # mate
        offspring.extend(toolbox.mate(ind1, ind2, optimize_resources))

    for mutant in offspring:
        if optimize_resources:
            # resource borders mutation
            toolbox.mutate_resource_borders(mutant)
        # other mutation
        toolbox.mutate(mutant)

    return offspring



from copy import deepcopy
from typing import Optional
from operator import itemgetter
from math import ceil

from sampo.schemas.graph import GraphNode, GraphEdge, WorkGraph, EdgeType
from sampo.schemas.works import WorkUnit
from sampo.schemas.requirements import WorkerReq, MaterialReq, ConstructionObjectReq, EquipmentReq

STAGE_SEP = '_stage_'


def make_new_node_id(work_unit_id: str, ind: int) -> str:
    """
    Creates an auxiliary id for restructuring the graph

    :param work_unit_id: str - id of work unit
    :param ind: int - sequence number of work_unit stage
    :return:
       auxiliary_id: str - an auxiliary id for the work unit
    """
    return f'{work_unit_id}{STAGE_SEP}{ind}'


def fill_parents_to_new_nodes(origin_node: GraphNode, id2new_nodes: dict[str, GraphNode],
                              restructuring_edges2new_nodes_id: dict[tuple[str, str], str],
                              use_lag_edge_optimization: bool):
    """
    Restores parent edges for transformed node

    :param origin_node: GraphNode - The original unconverted node
    :param id2new_nodes: GraphNodeDict - Dictionary with restructured new nodes id
        where the restructured nodes are written
    :param restructuring_edges2new_nodes_id: dict[tuple[str, str], str] - Dictionary for matching edges in the original
        work graph and IDs of new nodes that logically match those edges
    :param use_lag_edge_optimization: bool - if true - considers lags amount in edges,
        otherwise considers lags equal to zero and LagFinishStart edges as FinishStart edges
    :return: Nothing
    """
    # last stage id is equal to original node id
    last_stage_id = origin_node.id
    # make first stage id
    first_stage_id = make_new_node_id(origin_node.id, 0)
    # if first stage id not in id2new_nodes then there are no stages, so then first stage id is equal to last stage id
    first_stage_id = first_stage_id if first_stage_id in id2new_nodes else last_stage_id

    # basic time lag of FinishStart or InseparableFinishStart edges
    indent = 0

    # list of edges to parent nodes of first stage
    parents_first_stage: list[tuple[GraphNode, float, EdgeType]] = []
    # list of edges to parent nodes of last stage
    parents_last_stage: list[tuple[GraphNode, float, EdgeType]] = []
    # iterate over edges to parent nodes of original node
    for edge in origin_node.edges_to:
        # TODO Check indent application
        if edge.type in [EdgeType.FinishStart, EdgeType.InseparableFinishStart]:
            if edge.type is EdgeType.InseparableFinishStart:
                # if edge type is InseparableFinishStart then lag is equal to basic time lag
                lag = indent
            else:
                # if edge type is FinishStart then use time lag of original edge
                lag = edge.lag if not edge.lag % 1 else ceil(edge.lag)
                # lag = lag if lag > 0 else indent
            # edges of types FinishStart or InseparableFinishStart must enter the first stage node
            parents_first_stage.append((id2new_nodes[edge.start.id], lag, edge.type))
        elif not use_lag_edge_optimization:
            # if the use_lag_edge_optimization is False, then adding edges is trivial,
            # since a node can be divided into a maximum of three stages,
            # with the intermediate one not being used in creating edges
            match edge.type:
                case EdgeType.StartStart:
                    # make id of created first stage node of original parent node
                    new_parent_node_id = make_new_node_id(edge.start.id, 0)
                    # add edge between parent's first stage node and first stage of current node
                    parents_first_stage.append((id2new_nodes[new_parent_node_id], indent, EdgeType.FinishStart))
                case EdgeType.FinishFinish:
                    # add edge between parent's created last stage node and last stage of current node
                    parents_last_stage.append((id2new_nodes[edge.start.id], indent, EdgeType.FinishStart))
                case EdgeType.LagFinishStart:
                    # if the use_lag_edge_optimization is False,
                    # then edge of type LagFinishStart is just replaced by an edge of type FinishStart
                    # so, add edge between parent's created last stage node and first stage of current node
                    parents_first_stage.append((id2new_nodes[edge.start.id], indent, EdgeType.FinishStart))
        else:
            match edge.type:
                case EdgeType.StartStart:
                    # get id of corresponding created stage node of original parent node
                    new_parent_node_id = restructuring_edges2new_nodes_id.pop((edge.start.id, edge.finish.id))
                    # get parent's stage node by received id and add edge between it and first stage of current node
                    parents_first_stage.append((id2new_nodes[new_parent_node_id], indent, EdgeType.FinishStart))
                case EdgeType.FinishFinish:
                    # get id of corresponding stage of current node and get the stage node itself
                    stage_node = id2new_nodes[restructuring_edges2new_nodes_id.pop((edge.finish.id, edge.start.id))]
                    # add edge between parent's created last stage node and received stage node
                    stage_node.add_parents([(id2new_nodes[edge.start.id], indent, EdgeType.FinishStart)])
                case EdgeType.LagFinishStart:
                    # get id of corresponding created stage node of original parent node
                    new_parent_node_id = restructuring_edges2new_nodes_id.pop((edge.start.id, edge.finish.id))
                    # get parent's stage node by received id and add edge between it and first stage of current node
                    parents_first_stage.append((id2new_nodes[new_parent_node_id], indent, EdgeType.FinishStart))
                    # get id of corresponding stage of current node and get the stage node itself
                    stage_node = id2new_nodes[restructuring_edges2new_nodes_id.pop((edge.finish.id, edge.start.id))]
                    # add edge between parent's created last stage node and received stage node
                    stage_node.add_parents([(id2new_nodes[edge.start.id], indent, EdgeType.FinishStart)])

    # add parents to first stage of current node
    id2new_nodes[first_stage_id].add_parents(parents_first_stage)
    # add parents to last stage of current node
    id2new_nodes[last_stage_id].add_parents(parents_last_stage)


def split_node_into_stages(origin_node: GraphNode, restructuring_edges: list[tuple[GraphEdge, bool]],
                           id2new_nodes: dict[str, GraphNode],
                           restructuring_edges2new_nodes_id: dict[tuple[str, str], str],
                           use_lag_edge_optimization: bool):
    """
    Splits the node into stages according to the lags in the restructuring edges.
    For all stages except the last one, the id changes. For the last one, the id remains the same,
    so that it is more convenient to restore the edges.
    The resulting nodes are chained together by Inseparable-Finish-Start edges.

    :param restructuring_edges: list[tuple[GraphEdge, bool]] - list of restructuring edges and bool flag of reversion
    :param origin_node: GraphNode - Node to be divided into stages
    :param id2new_nodes: GraphNodeDict - Dictionary with restructured new nodes id where the restructured nodes will
        be written
    :param restructuring_edges2new_nodes_id: dict[tuple[str, str], str] - Dictionary for matching edges in the original
        work graph and IDs of new nodes that logically match those edges
    :param use_lag_edge_optimization: bool - if true - considers lags amount in edges,
        otherwise considers lags equal to zero and LagFinishStart edges as FinishStart edges
    :return: Nothing
    """

    def get_reqs_amounts(volume_proportion: float, reqs2amounts_accum: dict[str, list[int]]) \
            -> tuple[dict[str, list[int]], dict[str, list[int]]]:
        """
        Creates for the new stage node a mapper
        of work unit requirement names and amounts calculated based on volume_proportion.
        Updates the accumulated amounts in reqs2amounts_accum by adding calculated for the new stage node amounts
        """
        # make a mapper of requirement names and amounts for the new stage node
        reqs2amounts = {}
        for reqs in reqs2classes:
            # define a name of the amount attribute of the requirement
            attr = 'volume' if reqs == 'worker_reqs' else 'count'
            # calculate requirement amounts for the new stage node based on volume_proportion
            new_amounts = [int(volume_proportion * getattr(req, attr)) for req in getattr(wu, reqs)]
            reqs2amounts[reqs] = new_amounts
            # update the accumulated amounts of the requirement
            reqs2amounts_accum[reqs] = [accum_amount + amount
                                        for accum_amount, amount in zip(reqs2amounts_accum[reqs], new_amounts)]
        return reqs2amounts, reqs2amounts_accum

    def make_new_stage_node(volume_proportion: float,
                            edge_with_prev_stage_node: list[tuple[GraphNode, float, EdgeType]],
                            wu_attrs: dict,
                            reqs2attrs: dict[str, list[dict]]) -> GraphNode:
        """
        This function updates the requirements attributes of the original node with new pre-calculated amounts.
        Using these updated attributes, new requirement instances are created
        and the work unit attributes are updated accordingly. Also, in the attributes of the work unit, the id, name
        and volume are updated. Volume calculates based on volume_proportion.
        The resulting attributes are then used to create new Work Unit to make returned new stage Node.

        :param volume_proportion: proportion of volume for new stage
        :param edge_with_prev_stage_node: list with edge to previous stage node
        :param wu_attrs: copied attributes of work unit of original node
        :param reqs2attrs: copied attributes of requirements of original node

        :return: created Graph Node of new stage
        """
        # new requirements to update work unit attributes
        new_reqs = {}
        for reqs, req_class in reqs2classes.items():
            # define a name of the amount attribute of the requirement
            attr = 'volume' if reqs == 'worker_reqs' else 'count'
            for req_attrs, amount in zip(reqs2attrs[reqs], reqs2amounts[reqs]):
                # update amount attribute in requirement attributes
                req_attrs[attr] = amount
            # make new requirement instances
            new_reqs[reqs] = [req_class(**attrs) for attrs in reqs2attrs[reqs]]
        # update work unit attributes with created requirements
        wu_attrs.update(new_reqs)
        # update id attribute with current stage node id
        wu_attrs['id'] = stage_node_id
        # update name attribute with current index of stage
        wu_attrs['name'] = f'{wu.name}{STAGE_SEP}{stage_i}'
        # update volume attribute with passed proportion
        wu_attrs['volume'] = wu.volume * volume_proportion
        # make new work unit for new stage node with updated attributes
        new_wu = WorkUnit(**wu_attrs)
        # make new graph node for new stage with created work unit and with passed edge to previous stage node
        return GraphNode(new_wu, edge_with_prev_stage_node)

    def match_prev_edges_with_stage_nodes_id(restructuring_edges2new_nodes_id: dict[tuple[str, str], str]):
        """
        Matches the original edges, along which a previous stage node has been created,
        and IDs of the previous and current stage nodes.
        It is needed because the edges entering the node must refer to the next stage node
        for further filling of parent nodes, so matching can be done only after creating the next stage node.
        Edges that leave the node matches to the previous stage node.
        """
        for prev_edge, prev_is_edge_to_node in edges_to_match_with_stage_nodes:
            start, finish = prev_edge.start.id, prev_edge.finish.id
            if prev_is_edge_to_node:
                start, finish = finish, start
            restructuring_edges2new_nodes_id[(start, finish)] = \
                prev_stage_node_id if not prev_is_edge_to_node else stage_node_id

    wu = origin_node.work_unit
    # copy attributes of original work unit
    wu_attrs = dict(wu.__dict__)

    if not restructuring_edges:
        # if there are no edges dividing the node into stages, then simply copy this node
        id2new_nodes[wu.id] = GraphNode(deepcopy(wu), [])
        return

    # define mapper of requirements attribute names and classes
    reqs2classes = {'worker_reqs': WorkerReq, 'equipment_reqs': EquipmentReq,
                    'object_reqs': ConstructionObjectReq, 'material_reqs': MaterialReq}
    # make mapper of work unit requirement names and copied class object attributes
    reqs2attrs = {reqs: [dict(req.__dict__) for req in getattr(wu, reqs)] for reqs in reqs2classes}
    # make mapper of work unit requirement names and accumulated amounts
    reqs2amounts_accum = {reqs: [0 for _ in getattr(wu, reqs)] for reqs in reqs2classes}

    # calculate from the restructuring edges all the proportions of dividing the node into stages
    # depending on the direction and lags of these edges.
    # if an edge comes to a node,
    # then the proportion is calculated as 1 - lag / start.volume if use_lag_edge_optimization else 1
    # if an edge leaves a node,
    # then the proportion is calculated as lag / volume if use_lag_edge_optimization else 0
    proportions_accum = [(int(is_edge_to_node) +
                          ((1 - 2 * int(is_edge_to_node)) * edge.lag / edge.start.work_unit.volume
                           if use_lag_edge_optimization and 0 < edge.lag <= edge.start.work_unit.volume
                           else 0
                           ),
                          edge, is_edge_to_node
                          )
                         for edge, is_edge_to_node in restructuring_edges]
    # sort the resulting proportions in ascending order so that they represent a list of accumulations
    proportions_accum.sort(key=itemgetter(0))
    # initialize index of stages for creating name and id of stage nodes
    stage_i = 0

    # Create first stage node

    # get first proportion with corresponding edge
    accum, edge, is_edge_to_node = proportions_accum[0]
    # get amounts of requirements for the first stage node
    reqs2amounts, reqs2amounts_accum = get_reqs_amounts(accum, reqs2amounts_accum)
    # make id for first stage node
    stage_node_id = make_new_node_id(wu.id, stage_i)
    # make first stage node and add it to id2new_nodes
    id2new_nodes[stage_node_id] = make_new_stage_node(accum, [], wu_attrs, reqs2attrs)

    # initialize a list that stores the edges along which a stage node has already been created.
    # used only if the use_lag_edge_optimization is True,
    # since otherwise the matching of the edges and new node IDs is trivial
    edges_to_match_with_stage_nodes = [(edge, is_edge_to_node)] if use_lag_edge_optimization else None

    # Create intermediate stage nodes

    # iterate over the accumulations following the first, checking the difference with the previous accumulation
    for value, prev_value in zip(proportions_accum[1:], proportions_accum):
        accum, edge, is_edge_to_node = value
        prev_accum, _, _ = prev_value
        if accum == prev_accum:
            # the difference is zero, therefore the corresponding division into the stage has already been made
            if use_lag_edge_optimization:
                # if the use_lag_edge_optimization is True,
                # then add this edge for further matching with IDs of stage nodes
                # when the next stage node will be created
                edges_to_match_with_stage_nodes.append((edge, is_edge_to_node))
            continue
        # increase the stage index for creating name and id of stage node
        stage_i += 1
        # remember the ID of the previous stage node for matching with stored edges
        prev_stage_node_id = stage_node_id
        # make id for new stage node
        stage_node_id = make_new_node_id(wu.id, stage_i)
        # calculate difference between accumulations to obtain the proportion for splitting into the next stage
        proportion = accum - prev_accum
        # get amounts of requirements for the new stage node
        reqs2amounts, reqs2amounts_accum = get_reqs_amounts(proportion, reqs2amounts_accum)
        # make new stage node with InseparableFinishStart edge to previous stage node and add it to id2new_nodes
        id2new_nodes[stage_node_id] = make_new_stage_node(proportion, [(id2new_nodes[prev_stage_node_id], 0,
                                                                        EdgeType.InseparableFinishStart)],
                                                          wu_attrs, reqs2attrs
                                                          )
        if use_lag_edge_optimization:
            # if the use_lag_edge_optimization is True,
            # match stored edges with stage node IDs since a new stage node has been created
            match_prev_edges_with_stage_nodes_id(restructuring_edges2new_nodes_id)
            # store this edge for further matching with IDs of stage nodes
            # when the next stage node will be created
            edges_to_match_with_stage_nodes = [(edge, is_edge_to_node)]

    # Create last stage node

    # increase the stage index for name of last stage node
    stage_i += 1
    # remember the ID of the previous stage node for matching with stored edges
    prev_stage_node_id = stage_node_id
    # ID of the last stage node is equal to the ID of the original node,
    # so that it is more convenient to restore the parent edges
    stage_node_id = wu.id
    # proportion for the last stage is calculated as one minus the last accumulation
    proportion = 1 - accum
    # iterate over the requirements and calculate their amounts for the last stage
    for reqs in reqs2classes:
        # define a name of the amount attribute of the requirement
        attr = 'volume' if reqs == 'worker_reqs' else 'count'
        # amount of requirements for the last stage is calculated
        # as the difference between the amounts of the original node and the accumulated requirement amounts
        reqs2amounts[reqs] = [getattr(req, attr) - req_accum
                              for req, req_accum in zip(getattr(wu, reqs), reqs2amounts_accum[reqs])]
    # make last stage node with InseparableFinishStart edge to previous stage node and add it to id2new_nodes
    id2new_nodes[stage_node_id] = make_new_stage_node(proportion, [(id2new_nodes[prev_stage_node_id], 0,
                                                                    EdgeType.InseparableFinishStart)],
                                                      wu_attrs, reqs2attrs
                                                      )
    if use_lag_edge_optimization:
        # if the use_lag_edge_optimization is True,
        # match stored edges with stage node IDs since the last stage node has been created
        match_prev_edges_with_stage_nodes_id(restructuring_edges2new_nodes_id)


def graph_restructuring(wg: WorkGraph, use_lag_edge_optimization: Optional[bool] = False) -> WorkGraph:
    """
    Rebuilds all edges into Finish-Start and Inseparable-Finish-Start edges
    with the corresponding rebuilding of the nodes and returns new work graph

    :param wg: WorkGraph - The graph to be converted
    :param use_lag_edge_optimization: bool - if true - considers lags amount in edges,
        otherwise considers lags equal to zero and LagFinishStart edges as FinishStart edges
    :return:
        new_work_graph: WorkGraph - restructured graph
    """

    def get_restructuring_edges(edges: list[GraphEdge], edge_type: EdgeType, is_edge_to_node: bool) \
            -> list[tuple[GraphEdge, bool]]:
        """
        This function keeping all the edges of the types
        that need to be removed and which will break the node into stages
        and a flag indicating whether the edge enters or leaves the node
        """
        return [(edge, is_edge_to_node) for edge in edges if edge.type is edge_type]

    # mapper of IDs of new nodes and the new nodes themselves that represent the restructured graph
    id2new_nodes: dict[str, GraphNode] = dict()
    # mapper of original edges and IDs of new nodes that should be connected.
    # edge is specified by two IDs of the nodes that this edge connects.
    # the order of the node IDs determines (based on the first ID) on which side of the edge the new node ID is written.
    # used only if the use_lag_edge_optimization is True, since otherwise the connection is trivial
    restructuring_edges2new_nodes_id: dict[tuple[str, str], str] = dict()
    # iterate over each node,
    # keeping all the edges that need to be removed and which will break the node into stages,
    # creating a new node (possibly with stage nodes) that will be in the final restructured graph,
    # and connecting to parent nodes that have already been created since the nodes in wg.nodes are topological sorted
    for node in wg.nodes:
        restructuring_edges = get_restructuring_edges(node.edges_from, EdgeType.StartStart, False) + \
                              get_restructuring_edges(node.edges_to, EdgeType.FinishFinish, True)

        if use_lag_edge_optimization:
            restructuring_edges += get_restructuring_edges(node.edges_to, EdgeType.LagFinishStart, True) + \
                                   get_restructuring_edges(node.edges_from, EdgeType.LagFinishStart, False)

        split_node_into_stages(node, restructuring_edges, id2new_nodes, restructuring_edges2new_nodes_id,
                               use_lag_edge_optimization)
        fill_parents_to_new_nodes(node, id2new_nodes, restructuring_edges2new_nodes_id, use_lag_edge_optimization)

    return WorkGraph(id2new_nodes[wg.start.id], id2new_nodes[wg.finish.id])



from typing import FrozenSet

WorkerSpecialization = str

DRIVER: WorkerSpecialization = 'driver'
FITTER: WorkerSpecialization = 'fitter'
HANDYMAN: WorkerSpecialization = 'handyman'
ELECTRICIAN: WorkerSpecialization = 'electrician'
MANAGER: WorkerSpecialization = 'manager'
ENGINEER: WorkerSpecialization = 'engineer'

WORKER_TYPES: FrozenSet[WorkerSpecialization] = frozenset({DRIVER, FITTER, HANDYMAN, ELECTRICIAN, MANAGER, ENGINEER})



import heapq
import math
from abc import ABC, abstractmethod
from functools import cached_property

import numpy as np
from sortedcontainers import SortedList

from sampo.schemas.landscape_graph import LandGraph, LandGraphNode, LandEdge
from sampo.schemas.resources import Material
from sampo.schemas.zones import ZoneConfiguration


class ResourceSupply(ABC):
    def __init__(self, id: str, name: str):
        self.id = id
        self.name = name

    @abstractmethod
    def get_resources(self) -> list[tuple[str, int]]:
        ...


class Road(ResourceSupply):
    def __init__(self, name: str,
                 edge: LandEdge,
                 speed: float = 5):
        """
        :param name: name of road
        :param edge: the edge in LandGraph
        :poram bandwidth: the number of vehicles that road can pass per hour
        :param speed: the maximum value of speed on the road
        :param vehicles: the number of vehicles that are on the road at the current moment
        """
        super(Road, self).__init__(edge.id, name)
        self.vehicles = edge.bandwidth
        self.length = edge.weight
        self.speed = speed
        self.overcome_time = math.ceil(self.length / self.speed)
        self.edge = edge

    def get_resources(self) -> list[tuple[str, int]]:
        return [('speed', self.speed), ('length', self.edge.weight), ('vehicles', self.vehicles)]


class Vehicle(ResourceSupply):
    def __init__(self,
                 id: str,
                 name: str,
                 capacity: list[Material]):
        super(Vehicle, self).__init__(id, name)
        self.capacity = capacity
        self.cost = 0.0
        self.volume = sum([mat.count for mat in self.capacity])

    @cached_property
    def resources(self) -> dict[str, int]:
        return {mat.name: mat.count for mat in self.capacity}

    def get_resources(self) -> list[tuple[str, int]]:
        return [(mat.name, mat.count) for mat in self.capacity]


class ResourceHolder(ResourceSupply):
    def __init__(self,
                 id: str,
                 name: str,
                 vehicles: list[Vehicle] = None,
                 node: LandGraphNode = None):
        """
        :param name:
        :param vehicles:
        :param node:
        """
        # let ids of two objects will be similar to make simpler matching ResourceHolder to node in LandGraph
        super(ResourceHolder, self).__init__(id, name)
        self.node_id = node.id
        self.vehicles = vehicles
        self.node = node

    def get_resources(self) -> list[tuple[str, int]]:
        return [(name, count) for name, count in self.node.resource_storage_unit.capacity.items()]


class LandscapeConfiguration:
    def __init__(self,
                 holders: list[ResourceHolder] = None,
                 lg: LandGraph = None,
                 zone_config: ZoneConfiguration = ZoneConfiguration()):
        self.WAY_LENGTH = np.Inf
        self.dist_mx: list[list[float]] = None
        self.path_mx: np.array = None
        self.road_mx: list[list[str]] = None
        if holders is None:
            holders = []
        self.lg: LandGraph = lg
        self._holders: list[ResourceHolder] = holders

        # _ind2holder_id is required to match ResourceHolder's id to index in list of LangGraphNodes to work with routing_mx
        self.ind2holder_node_id: dict[int, str] = {self.lg.node2ind[holder.node]: holder.node.id for holder in
                                                   self._holders}
        self.holder_node_id2resource_holder: dict[str, ResourceHolder] = {holder.node.id: holder for holder in self._holders}
        self.zone_config = zone_config
        if self.lg is None:
            return
        self.works2platform = {}
        for platform in self.platforms:
            for work in platform.works:
                self.works2platform[work] = platform
        # self.works2platform: dict['GraphNode', LandGraphNode] = {work: platform for platform in self.platforms
        #                                                          for work in platform.works}
        self._build_routes()
        self._node2ind = self.lg.node2ind

    def get_sorted_holders(self, node: LandGraphNode) -> SortedList[list[tuple[float, str]]]:
        """
        :param node: id of node in LandGraph's list of nodes
        :return: sorted list of holders' id by the length of way
        """
        holders = []
        node_ind = self._node2ind[node]
        for i in self.ind2holder_node_id.keys():
            if self.dist_mx[node_ind][i] != self.WAY_LENGTH:
                holders.append((self.dist_mx[node_ind][i], self.ind2holder_node_id[i]))
        return SortedList(holders, key=lambda x: x[0])

    def get_route(self, from_node: LandGraphNode, to_node: LandGraphNode) -> list[str]:
        """
        Return a list of roads' id that are part of the route
        """
        from_ind = self._node2ind[from_node]
        to_ind = self._node2ind[to_node]

        path = [to_ind]
        to = to_ind
        while self.path_mx[from_ind][to] != from_ind:
            path.append(self.path_mx[from_ind][to])
            to = self.path_mx[from_ind][to]
        path.append(from_ind)

        return [self.road_mx[path[v - 1]][path[v]] for v in range(len(path) - 1, 0, -1)]

    def construct_route(self, from_node: LandGraphNode, to_node: LandGraphNode,
                        roads_available: list[Road]) -> list[str]:
        """
        Construct the route from the list of available roads whether it is possible.
        :param roads_available: list of available roads
        :return: list of roads' id that are included to the route whether it exists, otherwise return empty list
        """
        def dijkstra(node_ind: int):
            prior_queue = [(.0, node_ind)]
            distances[node_ind] = 0

            while prior_queue:
                dist, v = heapq.heappop(prior_queue)

                if dist > distances[v]:
                    continue

                for road in self.lg.nodes[v].roads:
                    if road.id not in roads_available_set:
                        continue
                    d = dist + road.weight
                    finish_ind = self._node2ind[road.finish]
                    if d < distances[finish_ind] and road.id in roads_available_set:
                        distances[finish_ind] = d
                        path_mx[node_ind][finish_ind] = v
                        heapq.heappush(prior_queue, (d, finish_ind))

        distances = [self.WAY_LENGTH] * self.lg.vertex_count
        path_mx = np.full((self.lg.vertex_count, self.lg.vertex_count), -1)
        roads_available_set = set(road.id for road in roads_available)
        for v in range(self.lg.vertex_count):
            for u in range(self.lg.vertex_count):
                if v == u:
                    path_mx[v][u] = 0
                elif self.lg.adj_matrix[v][u] != np.Inf and self.road_mx[v][u] in roads_available_set:
                    path_mx[v][u] = v

        from_ind = self._node2ind[from_node]
        to_ind = self._node2ind[to_node]
        dijkstra(to_ind)

        if path_mx[to_ind][from_ind] == -1:
            return []

        path_tmp = []
        fr = from_ind
        while path_mx[to_ind][fr] != to_ind:
            path_tmp.append(path_mx[to_ind][fr])
            fr = path_mx[to_ind][fr]
        path = [from_ind]
        path.extend(path_tmp)
        path.append(to_ind)
        return [self.road_mx[path[v]][path[v + 1]] for v in range(len(path) - 1)]

    @cached_property
    def holders(self) -> list[ResourceHolder]:
        return self._holders

    @cached_property
    def platforms(self) -> list[LandGraphNode]:
        if self.lg is None:
            return []
        platform_ids = set([node.id for node in self.lg.nodes]).difference(
            set([holder.node_id for holder in self._holders]))
        return [node for node in self.lg.nodes if node.id in platform_ids]

    @cached_property
    def roads(self) -> list[Road]:
        return [Road(f'road_{i}', edge) for i, edge in enumerate(self.lg.edges)]

    def get_all_resources(self) -> list[dict[str, dict[str, int]]]:
        def merge_dicts(a, b):
            c = a.copy()
            c.update(b)
            return c

        if self.lg is None:
            return []

        holders = {
            holder.id: merge_dicts({name: count for name, count in holder.get_resources()},
                                   {'vehicles': len(holder.vehicles)})
            for holder in self.holders}
        roads = {road.id: {'vehicles': road.vehicles}
                 for road in self.roads}

        resources = [holders, roads]
        return resources

    def get_platforms_resources(self) -> dict[str, dict[str, int]]:
        return {node.id: {name: count for name, count in node.resource_storage_unit.capacity.items()}
                for node in self.platforms}

    def _build_routes(self):
        count = self.lg.vertex_count
        dist_mx = self.lg.adj_matrix.copy()
        path_mx: np.array = np.full((count, count), -1)
        road_mx: list[list[str]] = [['-1' for j in range(count)] for i in range(count)]

        for v in range(count):
            for u in range(count):
                if v == u:
                    path_mx[v][u] = 0
                elif dist_mx[v][u] != np.Inf:
                    path_mx[v][u] = v
                    for road in self.lg.nodes[v].roads:
                        if self.lg.node2ind[road.finish] == u:
                            road_mx[v][u] = road.id
                else:
                    path_mx[v][u] = -1

        for i in range(self.lg.vertex_count):
            for u in range(self.lg.vertex_count):
                for v in range(self.lg.vertex_count):
                    if (dist_mx[u][i] != np.Inf and dist_mx[u][i] != 0
                            and dist_mx[i][v] != np.Inf and dist_mx[i][v] != 0
                            and dist_mx[u][i] + dist_mx[i][v] < dist_mx[u][v]):
                        dist_mx[u][v] = dist_mx[u][i] + dist_mx[i][v]
                        path_mx[u][v] = path_mx[i][v]

        self.dist_mx = dist_mx
        self.path_mx = path_mx
        self.road_mx = road_mx


class MaterialDelivery:
    def __init__(self, work_id: str):
        self.id = work_id
        self.delivery = {}

    def add_delivery(self, name: str, count: int, start_time: 'Time', finish_time: 'Time', from_holder: str):
        material_delivery = self.delivery.get(name, None)
        if material_delivery is None:
            material_delivery = []
            self.delivery[name] = material_delivery
        material_delivery.append((count, start_time, finish_time, from_holder))

    def add_deliveries(self, name: str, deliveries: list[tuple[int, 'Time', 'Time', str]]):
        material_delivery = self.delivery.get(name, None)
        if material_delivery is None:
            self.delivery[name] = deliveries
        else:
            material_delivery.extend(deliveries)


def max_fill(mat_max: int, mat_available: int):
    return mat_max - mat_available


def necessary_fill(mat_count: int, mat_available: int, mat_max: int):
    return mat_count + mat_available - mat_max



from abc import ABC, abstractmethod
from typing import Callable, Optional

import numpy as np

from sampo.scheduler.utils import WorkerContractorPool
from sampo.schemas.resources import Worker
from sampo.schemas.time import Time


class ResourceOptimizer(ABC):
    """
    Base class to build different methods of resource optimization.
    Constructed methods minimize the quantity of resources.
    """

    @abstractmethod
    def optimize_resources(self,
                           worker_pool: WorkerContractorPool,
                           worker_team: list[Worker],
                           optimize_array: Optional[np.ndarray],
                           down_border: np.ndarray,
                           up_border: np.ndarray,
                           get_finish_time: Callable[[list[Worker]], Time]):
        """
        The resource optimization module. Optimizes `worker_team` using `get_finish_time` metric.
        Should optimize `worker_team` in-place.

        :param worker_pool: global resources pool
        :param worker_team: worker team to optimize
        :param optimize_array: a boolean array that says what positions should be optimized
        :param down_border: down border of optimization
        :param up_border: up border of optimization
        :param get_finish_time: optimization function that should give execution time based on worker team
        """
        ...



from sampo.userinput.parser.contractor_type import ContractorType
from sampo.userinput.parser.csv_parser import CSVParser
from sampo.userinput.parser.exception import InputDataException, WorkGraphBuildingException



from abc import ABC, abstractmethod
from collections import defaultdict
from copy import deepcopy
from random import Random
from uuid import uuid4

from sampo.schemas import GraphNode, WorkGraph, WorkUnit, WorkTimeEstimator, IntervalGaussian, Interval
from sampo.schemas.time_estimator import DefaultWorkEstimator
from sampo.structurator.prepare_wg_copy import prepare_work_graph_copy


class StructureGenerationEstimator(ABC):
    @abstractmethod
    def generate_probabilities(self, parent: WorkUnit) -> dict[str, float]:
        ...

    @abstractmethod
    def get_volume(self, parent: WorkUnit, target: str) -> float:
        ...

class DefaultStructureGenerationEstimator(StructureGenerationEstimator):

    def __init__(self, rand: Random | None = None):
        self._gen_probabilities = defaultdict(lambda: defaultdict(float))
        self._rand = rand or Random()
        self._volume = IntervalGaussian(5, 2, 0.1, 10, self._rand)

    def set_probabilities(self, defect_probabilities: dict[str, dict[str, float]]) -> 'DefaultStructureGenerationEstimator':
        self._gen_probabilities = defect_probabilities
        return self

    def set_probability(self, parent: str, child: str, probability: float) -> 'DefaultStructureGenerationEstimator':
        self._gen_probabilities[parent][child] = probability
        return self

    def set_volume_interval(self, volume: Interval) -> 'DefaultStructureGenerationEstimator':
        self._volume = volume
        return self

    def generate_probabilities(self, parent: WorkUnit) -> dict[str, float]:
        return self._gen_probabilities[parent.name]

    def get_volume(self, parent: WorkUnit, target: str) -> float:
        return self._volume.rand_float()


class StructureEstimator(ABC):

    @abstractmethod
    def restruct(self, wg: WorkGraph) -> WorkGraph:
        ...

class DefaultStructureEstimator(StructureEstimator):

    def __init__(self, generator: StructureGenerationEstimator, rand: Random | None = None):
        self._generator = generator
        self._work_estimator = DefaultWorkEstimator()
        self._rand = rand or Random()

    def set_work_estimator(self, work_estimator: WorkTimeEstimator) -> 'DefaultStructureEstimator':
        self._work_estimator = work_estimator
        return self

    def _construct_sub_work(self, work_unit: WorkUnit, sub_work_name: str) -> WorkUnit:
        zone_reqs = deepcopy(work_unit.zone_reqs)
        volume = float(self._generator.get_volume(work_unit, sub_work_name))

        worker_reqs = self._work_estimator.find_work_resources(sub_work_name, volume)
        return WorkUnit(str(uuid4()), sub_work_name,
                        worker_reqs=worker_reqs, zone_reqs=zone_reqs, volume=volume)

    def _gen_sub_works(self, work_unit: WorkUnit) -> list[WorkUnit]:
        probabilities = list(self._generator.generate_probabilities(work_unit).items())
        return [self._construct_sub_work(work_unit, name)
                for name, prob in probabilities
                if self._rand.random() < prob]

    def restruct(self, wg: WorkGraph) -> WorkGraph:
        nodes, _ = prepare_work_graph_copy(wg, change_id=False)
        start = nodes[wg.start.id]
        finish = nodes[wg.finish.id]
        for i, node in enumerate(nodes.values()):
            if node.work_unit.is_service_unit:
                continue
            for sub_work in self._gen_sub_works(node.work_unit):
                # construct node, add edge to the parent
                sub_node = GraphNode(sub_work, [node])
                # add edge to the children of the routine
                finish.add_parents([sub_node])

        return WorkGraph(start, finish)



from sampo.scheduler.genetic.base import GeneticScheduler
from sampo.scheduler.genetic.converter import ScheduleGenerationScheme
from sampo.scheduler.genetic.operators import (TimeFitness, SumOfResourcesPeaksFitness, SumOfResourcesFitness,
                                               TimeWithResourcesFitness, DeadlineResourcesFitness, DeadlineCostFitness,
                                               TimeAndResourcesFitness)



from enum import Enum

class PipelineError(Exception):
    """
    Raised when any pipeline error occurred.

    This is a kind of 'IllegalStateException', e.g. raising this
    indicates that the corresponding pipeline come to incorrect internal state.
    """
    def __init__(self, message: str):
        super().__init__(message)

from sampo.pipeline.base import InputPipeline
from sampo.pipeline.default import DefaultInputPipeline


class PipelineType(Enum):
    DEFAULT = 0


class SchedulingPipeline:
    @staticmethod
    def create(pipeline_type: PipelineType = PipelineType.DEFAULT) -> InputPipeline:
        match pipeline_type:
            case PipelineType.DEFAULT:
                return DefaultInputPipeline()
            case _:
                raise PipelineError('Unknown pipeline type')
